from collections.abc import Callable
from enum import IntEnum
from typing import Any, Generic, ParamSpec, TypeVar

_P = ParamSpec("_P")
_T = TypeVar("_T")

class JSError(Exception):
    """Represents Numerous Exceptions raised from CYJS"""

    ...

class MemoryUsage:
    malloc_size: int
    malloc_limit: int
    memory_used_size: int
    malloc_count: int
    memory_used_count: int
    atom_count: int
    atom_size: int
    str_count: int
    str_size: int
    obj_count: int
    obj_size: int
    prop_count: int
    prop_size: int
    shape_count: int
    shape_size: int
    js_func_count: int
    js_func_size: int
    js_func_code_size: int
    js_func_pc2line_count: int
    array_count: int
    fast_array_count: int
    fast_array_elements: int
    binary_object_count: int
    binary_object_size: int

class PromiseHookType(IntEnum):
    INIT = 0
    BEFORE = 1
    AFTER = 2
    RESOLVE = 3

INIT: PromiseHookType = ...
BEFORE: PromiseHookType = ...
AFTER: PromiseHookType = ...
RESOLVE: PromiseHookType = ...

class Runtime:
    def __init__(self) -> None: ...
    def compute_memory_usage(self) -> Any: ...
    def dump_memory_usage(self, file: object) -> object: ...
    def execute_pending_job(self) -> object: ...
    def is_job_pending(self) -> bool: ...
    def run_gc(self) -> Any:
        """
        Runs QuickJS-NG's internal Garbage Collector

        **Warning!** Use at your own risk.
        """
        ...

    def set_memory_limit(self, limit: int) -> Any: ...
    def set_max_stack_size(self, max_stack_size: int) -> Any: ...
    def update_statck_top(self) -> Any: ...
    def set_promise_hook(
        self,
        func: Callable[["Context", PromiseHookType, Promise, Promise | None], None],
    ) -> object: ...

class _OView:
    def __init__(self, obj: Object) -> None: ...
    def __len__(self):  # -> unsigned int:
        ...

class _ObjectItemsView(_OView):
    def __iter__(self):  # -> Generator[tuple[object, object], None, None]:
        ...

class _ObjectKeysView(_OView):
    def __iter__(self):  # -> Generator[object, None, None]:
        ...
    def __contains__(self, key: object):  # -> bool:
        ...

class _ObjectValuesView(_OView):
    def __iter__(self):  # -> Generator[object, None, None]:
        ...
    def __contains__(self, value: object):  # -> bool:
        ...

class Object:
    def to_json(self) -> bytes:
        """
        Useful when debugging or handling unknown js to py conversions
        NOTE: for best results, using third party libraries like orjson
        or msgspec are advised.
        """
        ...

    @property
    def tag(self):  # -> signed int:
        ...
    def get(self, key):  # -> object:
        ...
    def set(self, key: object, value: object):  # -> None:
        ...
    def __call__(self, *args):  # -> object:
        ...
    def invoke(self, func: object, *args):  # -> object:
        ...
    def items(self) -> _ObjectItemsView: ...
    def values(self) -> _ObjectValuesView: ...
    def keys(self) -> _ObjectKeysView: ...

    # equivlent to context.eval_this(self)
    def eval(
        self,
        code: str | bytes | bytearray | memoryview,
        filename: object = ...,
        strict: bool = ...,
        backtrace_barrier: bool = ...,
        promise: bool = ...,
    ) -> object:
        """evaluates javascript code"""
        ...

    def eval_module(
        self,
        code: str | bytes | bytearray | memoryview,
        filename: object = ...,
        strict: bool = ...,
        backtrace_barrier: bool = ...,
        promise: bool = ...,
    ) -> object:
        """evaluates javascript module code"""
        ...

class JSFunction(Generic[_P, _T]):
    context: Context

    def __call__(self, *args: _P.args, **kwargs: _P.kwargs) -> _T: ...
    @property
    def object(self) -> Object:
        pass

class Context:
    def __init__(
        self,
        runtime: Runtime = ...,
        base_objects: bool = ...,
        date: bool = ...,
        intrinsic_eval: bool = ...,
        regexp_compiler: bool = ...,
        regexp: bool = ...,
        json: bool = ...,
        proxy: bool = ...,
        map_set: bool = ...,
        typed_arrays: bool = ...,
        bigint: bool = ...,
        weak_ref: bool = ...,
        performance: bool = ...,
        dom_exception: bool = ...,
        promise: bool = ...,
    ) -> None: ...
    def eval(
        self,
        code: str | bytes,
        filename: object = ...,
        strict: bool = ...,
        backtrace_barrier: bool = ...,
        promise: bool = ...,
    ) -> object:
        """evaluates javascript code"""
        ...

    def eval_module(
        self,
        code: str | bytes,
        filename: object = ...,
        strict: bool = ...,
        backtrace_barrier: bool = ...,
        promise: bool = ...,
    ) -> object:
        """evaluates javascript module code"""
        ...

    def get_global(self) -> Object: ...
    def json_parse(self, json: object) -> Object: ...
    def get(self, name: bytes | str) -> Object | Any:
        """Implements a Shortcut for converting a global object to a python
        object and setting a value to utilize
        off of."""
        ...

    def set(
        self, name: bytes | str, item: object | Object | Promise | JSFunction
    ) -> None:
        """Sets an item to the current globalThis object"""
        ...

    def add_function(
        self, func: Callable[_P, _T], name: str | bytes | None = None, magic: int = 11
    ) -> JSFunction[_P, _T]:
        """adds a python function to quickjs using
        `JS_NewCClosure` since Quickjs-ng doesn't have
        a good way for bidning python fucntions well yet

        :param func: the python function to invoke with
            quickjs note: that it may not pass along keyword arguments `**kw`
        :param name: an alternative name to give to the function being passed
        :param magic: the magic value of the js function (Not much is known about it
            at the moment... defaults to 11 which reflects quickjs's own tests)
        """

    def eval_this(
        self,
        code: str | bytes | bytearray | memoryview,
        this: Object | Any,
        filename: object = ...,
        strict: bool = ...,
        backtrace_barrier: bool = ...,
        promise: bool = ...,
    ) -> Any: ...
    def eval_this_with_module(
        self,
        code: str | bytes | bytearray | memoryview,
        this: Object | Any,
        filename: object = ...,
        strict: bool = ...,
        backtrace_barrier: bool = ...,
        promise: bool = ...,
    ) -> Any: ...

class CancelledError(Exception):
    """Promise was rejected"""

    ...

class InvalidStateError(Exception):
    """Promise on inavlid state"""

    ...

class Promise(Object):
    def add_done_callback(self, fn: Callable[["Promise"], None]) -> object:
        """Attaches a callable callback when promise finishes or raises an exception"""
        ...

    def exception(self) -> BaseException: ...
    def done(self) -> bool: ...
    def remove_done_callback(self, fn: Callable[["Promise"], None]) -> int:
        """Remove all instances of a callback from the "call when done" list.

        Returns the number of callbacks removed.
        """
        ...

    def result(self) -> Any: ...
    def poll(self) -> Any:
        """Polls QuickJS Eventloop a single cycle while attempting
        to wait for this Promise to complete"""
        ...



from cyjs._cyjs import Context

import pytest


# oldest trick in the book...
def add(a: int, b: int) -> int:
    return a + b


# TODO:
# def peform_raise():
#     raise RuntimeError("I'm a teapot")


def test_cclosure_eval(ctx: Context) -> None:
    func = ctx.add_function(add, "add")
    ctx.set("add", func)
    result = ctx.eval("globalThis.add(1, 2)")
    assert result == 3


def throw_exception():
    raise RuntimeError("Boo")


def test_function_that_raises_exception(ctx: Context):
    func = ctx.add_function(throw_exception, "py_throw_exception")
    ctx.set("py_throw", func)

    # The lower level gets the worm in this case since python does better with python
    # and Javascript does better with Javascript.
    with pytest.raises(RuntimeError, match=r"Boo"):
        ctx.eval("globalThis.py_throw()")



from cyjs import PromiseHookType, Context, Runtime, Promise
from dataclasses import dataclass
import pytest


# Reflects test api-test.c
# void promise_hook(void) Line: 523 - 661
# which contains pretty easy code to replicate.
# we don't run the entire thing but we want to at least ensure the callbacks are indeed 1 to 1.


@dataclass
class Counter:
    init: int = 0
    before: int = 0
    after: int = 0
    resolve: int = 0

    def on_promise(
        self,
        ctx: Context,
        hook: PromiseHookType,
        promise: Promise,
        parent: Promise | None,
    ):
        assert isinstance(promise, Promise), "promise was not actually a promise :("
        match hook:
            case PromiseHookType.INIT:
                self.init += 1
            case PromiseHookType.BEFORE:
                self.before += 1
            case PromiseHookType.AFTER:
                self.after += 1
            case PromiseHookType.RESOLVE:
                self.resolve += 1

    def is_equal_to(self, init: int, before: int, after: int, resolve: int) -> bool:
        return (
            self.init == init
            and self.before == before
            and self.after == after
            and self.resolve == resolve
        )


@dataclass
class Case:
    code: str
    expected: tuple[int, int, int, int]
    id: int

    def __str__(self) -> str:
        return str(self.id)


@pytest.fixture(scope="function")
def setup_ctx(rt: Runtime) -> tuple[Context, Counter]:
    ctx = Context(rt)
    counter = Counter()
    rt.set_promise_hook(counter.on_promise)
    return ctx, counter


@pytest.fixture(
    params=(
        Case("new Promise(() => {})", (3, 0, 0, 2), 1),
        Case("new Promise((resolve,reject) => resolve())", (3, 0, 0, 3), 2),
        Case("new Promise((resolve,reject) => reject())", (3, 0, 0, 2), 3),
        # Case("new Promise(resolve => resolve({then(resolve){ resolve() }}))", ()) TODO:...
    ),
    ids=str,
)
def promise_hook(request: pytest.FixtureRequest) -> Case:
    return request.param


def test_hooks(setup_ctx: tuple[Context, Counter], promise_hook: Case):
    ctx, counter = setup_ctx
    ret = ctx.eval_module(promise_hook.code)
    assert isinstance(ret, Promise)
    assert counter.is_equal_to(*promise_hook.expected)


def test_bad_hook(rt: Runtime) -> None:
    # Stress testing bad case scenarios
    with pytest.raises(TypeError):
        rt.set_promise_hook(None)


def callback_failure(*args):
    raise RuntimeError("Jumpscare!")


def test_callback_failure(rt: Runtime) -> None:
    ctx = Context(rt)
    rt.set_promise_hook(callback_failure)

    with pytest.raises(RuntimeError):
        ctx.eval_module("new Promise(() => {})")



from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext
import os
from pathlib import Path
import sys

use_system_lib = bool(int(os.environ.get("QUICKJS_USE_SYSTEM_LIB", 0)))


QUICKJS_DIR = Path("quickjs")

quickjs_sources = list(
    map(
        str,
        [
            QUICKJS_DIR / "cutils.c",
            QUICKJS_DIR / "dtoa.c",
            QUICKJS_DIR / "libregexp.c",
            QUICKJS_DIR / "libunicode.c",
            QUICKJS_DIR / "quickjs.c",
        ],
    )
)

DEFINE_MACROS = (
    [("WIN32_LEAN_AND_MEAN", "1"), ("_WIN32_WINNT", "0x0601")]
    if sys.platform == "win32"
    else []
)
EXTRA_COMPILE_ARGS = (
    ["/std:c11", "/experimental:c11atomics"] if sys.platform == "win32" else []
)

# Added so that things can compile a bit faster (hopefully)
EXTRA_IGNORE_COMPILE_ARGS = (
    [
        "/wd4018",  # -Wno-sign-conversion
        "/wd4061",  # -Wno-implicit-fallthrough
        "/wd4100",  # -Wno-unused-parameter
        "/wd4200",  # -Wno-zero-length-array
        "/wd4242",  # -Wno-shorten-64-to-32
        "/wd4244",  # -Wno-shorten-64-to-32
        "/wd4245",  # -Wno-sign-compare
        "/wd4267",  # -Wno-shorten-64-to-32
        "/wd4388",  # -Wno-sign-compare
        "/wd4389",  # -Wno-sign-compare
        "/wd4456",  # Hides previous local declaration
        "/wd4457",  # Hides function parameter
        "/wd4710",  # Function not inlined
        "/wd4711",  # Function was inlined
        "/wd4820",  # Padding added after construct
        "/wd4996",  # -Wdeprecated-declarations
        "/wd5045",  # Compiler will insert Spectre mitigation for memory load if /Qspectre switch specified
    ]
    if sys.platform == "win32"
    else [
        # IDK How to do all these flags yet but I am reviewing CMakeLists.txt in quickjs for clues...
        "-Wno-implicit-fallthrough",
        "-Wno-sign-compare",
        "-Wno-missing-field-initializers",
        "-Wno-unused-parameter",
        "-Wno-unused-but-set-variable",
        "-Wno-unused-result",
        "-Wno-stringop-truncation",
        "-Wno-array-bounds",
    ]
)

EXTRA_COMPILE_ARGS += EXTRA_IGNORE_COMPILE_ARGS


class quickjs_build_ext(build_ext):
    # Brought over from winloop since these can be very useful.

    user_options = build_ext.user_options + [
        ("cython-always", None, "run cythonize() even if .c files are present"),
        (
            "cython-annotate",
            None,
            "Produce a colorized HTML version of the Cython source.",
        ),
        ("cython-directives=", None, "Cythion compiler directives"),
    ]

    def initialize_options(self):
        self.cython_always = False
        self.cython_annotate = False
        self.cython_directives = None
        self.parallel = True
        super().initialize_options()

    # Copied from winloop
    def finalize_options(self):
        need_cythonize = self.cython_always
        cfiles = {}

        for extension in self.distribution.ext_modules:
            for i, sfile in enumerate(extension.sources):
                if sfile.endswith(".pyx"):
                    prefix, _ = os.path.splitext(sfile)
                    cfile = prefix + ".c"

                    if os.path.exists(cfile) and not self.cython_always:
                        extension.sources[i] = cfile
                    else:
                        if os.path.exists(cfile):
                            cfiles[cfile] = os.path.getmtime(cfile)
                        else:
                            cfiles[cfile] = 0
                        need_cythonize = True

        # from winloop & cyares
        if need_cythonize:
            # import pkg_resources

            # Double check Cython presence in case setup_requires
            # didn't go into effect (most likely because someone
            # imported Cython before setup_requires injected the
            # correct egg into sys.path.
            try:
                import Cython  # type: ignore  # noqa: F401
            except ImportError:
                raise RuntimeError("please install cython to compile cyjs from source")

            from Cython.Build import cythonize

            directives = {}
            if self.cython_directives:
                for directive in self.cython_directives.split(","):
                    k, _, v = directive.partition("=")
                    if v.lower() == "false":
                        v = False
                    if v.lower() == "true":
                        v = True
                    directives[k] = v
                self.cython_directives = directives

            self.distribution.ext_modules[:] = cythonize(
                self.distribution.ext_modules,
                compiler_directives=directives,
                annotate=self.cython_annotate,
                emit_linenums=self.debug,
                # Try using a cache to help with compiling as well...
                cache=True,
            )

        return super().finalize_options()


def pyx_ext(file: str):
    return Extension(
        f"cyjs.{file}",
        [f"cyjs/{file}.pyx"] + quickjs_sources,
        include_dirs=["quickjs"],
        define_macros=DEFINE_MACROS,
        # NOTE: You will need to fix stdalign.h like I did if yours didn't exist - Vizonex
        extra_compile_args=EXTRA_COMPILE_ARGS,
    )


if __name__ == "__main__":
    setup(
        ext_modules=[
            pyx_ext("_cyjs"),
        ],
        cmdclass={"build_ext": quickjs_build_ext},
    )



from ._cyjs import (
    CancelledError,
    Context,
    JSError,
    JSFunction,
    Object,
    Promise,
    PromiseHookType,
    Runtime,
)

__author__ = "Vizonex"
__version__ = "0.1.0"

__all__ = (
    "CancelledError",
    "Context",
    "JSError",
    "JSFunction",
    "Object",
    "Promise",
    "PromiseHookType",
    "Runtime",
)
